import datetime

from . import TypeModel
from .events import Event
from .rooms import Room


class Booking(TypeModel):
    base_path = "/booking/"
    def __init__(
            self, 
            event: Event,
            room: Room,
            time_start: datetime.datetime,
            time_finish: datetime.datetime, 
            id: int | None = None,
            additional_data: dict| None = None):
        self.id = id
        self.event = event
        self.room = room
        self.time_start = time_start
        self.time_finish = time_finish
        self.additional_data = additional_data

    @classmethod
    def from_json(cls, booking: dict):
        pass

    @property
    def body(self):
        """
            It can cause error when room or event didn't save in the database
        """
        return dict(
            time_start=self.time_start,
            time_finish=self.time_finish, 
            additional_data=self.additional_data,
            room_id=self.room.id,
            event_id=self.event.id)



from . import TypeModel


class Seat(TypeModel):
    base_path = "/seats/"

    def __init__(
            self,
            row: int,
            column: int,
            number: int,
            booked: bool = False,
            additional_data: dict | None = None):

        self.row = row
        self.column = column
        self.number = number
        self.booked = booked
        self.additional_data = additional_data

    @classmethod
    def from_json(cls, seat: dict):
        """
            Getting seat using json that we get from API
        """
        return cls(
            row=seat["row"], 
            column=seat["column"], 
            number=seat["number"], 
            booked=seat["booked"], 
            additional_data=seat["additional_data"]
        )

    @property
    def room(self):
        """
            TODO: Getting room by it id in the database and return it
        """
        pass

    def book(self):
        """
            TODO: booking seat in the database
        """
        if self.booked:
            raise ValueError("The seat have already booked")
        self.booked = True
        pass

    def unbook(self, strict: bool = False):
        """
            TODO: unbooking seat in the database
        """
        self.booked = False
        pass

    @property
    def body(self) -> dict:
        return dict(
            row=self.row,
            column=self.column, 
            number=self.number, 
            booked=self.booked, 
            additional_data=self.additional_data
        )

    @property
    def params(self) -> dict:
        return {}


from . import TypeModel


class Event(TypeModel):
    base_path = "/events/"
    
    def __init__(
            self,
            title: str,
            id: int | None = None,
            additional_data: dict | None = None):
        self.id = id
        self.title = title
        self.additional_data = additional_data

    @classmethod
    def from_json(cls, event: dict):
        """
            Getting event using json that we get from API
            {
                "title": "str",
                "additional_data": "dict"
            }
        """
        return cls(title=event["title"], additional_data=event["additional_data"])

    @property
    def body(self) -> dict:
        return dict(title=self.title, additional_data=self.additional_data)

    @property
    def params(self) -> dict:
        return {}


from abc import ABC, abstractclassmethod,abstractproperty


class TypeModel(ABC):

    @abstractclassmethod
    def from_json(object_json: dict):
        raise NotImplementedError
    
    @abstractproperty
    def body():
        raise NotImplementedError
    
    @abstractproperty
    def params():
        raise NotImplementedError


import requests
from typing import Literal

from .types import TypeModel
from .types.rooms import Room
from .types.events import Event
from .types.booking import Booking


class BookSystemSDK:
    def __init__(
            self,
            api_url: str,
            rooms: list[Room] | None = None,
            events: list[Event] | None = None,
            booking: list[Booking] | None = None):
        self.api_url = api_url
        self._rooms = rooms
        self._events = events
        self._booking = booking
    
    @property
    def rooms(self) -> list[Room]:
        if not self._rooms:
            url = f"{self.api_url}/rooms/"
            self._rooms = [Room.from_json(room) for room in self._make_request(url=url, method="GET")]
        return self._rooms

    @property
    def events(self) -> list[Event]:
        if not self._events:
            url = f"{self.api_url}/events/"
            self._events = [Event.from_json(event) for event in self._make_request(url=url, method="GET")]
        return self._events
    
    @property
    def booking(self) -> list[Booking]:
        if not self._booking:
            url = f"{self.api_url}/booking/"
            self._booking = [Booking.from_json(booking) for booking in self._make_request(url=url, method="GET")]
        return self._booking

    def _make_request(
            self,
            url: str,
            method: Literal["GET", "POST", "PATCH", "DELETE"],
            body: dict | None = None,
            params: dict | None = None) -> dict:
        response = requests.request(method=method, url=url, json=body, params=params)
        json = response.json()
        if response.status_code not in [200, 201, 204]:
            raise ValueError(json["detail"])
        return json
    
    def create(self, obj: TypeModel) -> TypeModel:
        url = f"{self.api_url}{obj.base_path}"
        return obj.from_json(self._make_request(url=url, method="POST", body=obj.body, params=obj.params))
    
    def refresh(self, obj: TypeModel) -> TypeModel:
        url = f"{self.api_url}{obj.base_path}{obj.id}"
        return obj.from_json(self._make_request(url=url, method="PATCH", body=obj.body, params=obj.params))

    def delete(self, obj: TypeModel | list[TypeModel]) -> None:
        url = f"{self.api_url}{obj.base_path}{obj.id}"
        self._make_request(url=url, method="DELETE")

    def get(self, model, **kwargs) -> TypeModel:
        url = f"{self.api_url}{model.base_path}"
        json = self._make_request(url=url, method="GET", params=kwargs)
        return model.from_json(json)